from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, ClassVar

if TYPE_CHECKING: # pragma: nocover
    from .details._base import VinDetails


def constant_info(info):
    """Emulate details logic to always return the same information."""
    return lambda details: type("", (), {"get": (lambda code: info)})


def candidate_by_year_model_mapping(mapping: dict[str, dict[str, str]], years: list[int]):
    candidate_mapping = {}
    for model_year_range, candidates in mapping.items():
        start_model_year, end_model_year = model_year_range.split('-')

        if not start_model_year and not end_model_year:
            candidate_mapping.update(candidates)
            continue

        start_model_year = int(start_model_year)
        end_model_year = (
            int(end_model_year)
            if end_model_year else
            max(datetime.now(tz=timezone.utc).year, start_model_year) + 1
        )

        filter_years = [year for year in years if start_model_year <= year <= end_model_year]
        if filter_years:
            candidate_mapping.update(candidates)

    return candidate_mapping


class Annotatable:

    annotate_titles: ClassVar = {}

    def annotate(self) -> dict[str, Any]:

        annotations = {}
        no_attr = set()

        for attr_name, label in self.annotate_titles.items():
            value = getattr(self, attr_name, no_attr)

            if value is no_attr:
                continue

            if isinstance(value, list):
                value = ', '.join(f'{val}' for val in value)

            annotations[label] = f'{value}'

        return dict(sorted(annotations.items(), key=lambda item: item[0]))


class Assembler:
    """Assembler is a manufacturer that has a WMI and assemble vehicles for other brands using its own WMI."""
    __slots__ = ['manufacturer']

    brands: ClassVar[set['Brand']] = None

    def __init__(self, manufacturer: str | None = None):
        self.manufacturer = manufacturer or self.title

    @property
    def title(self) -> str:
        return self.__class__.__name__

    def __str__(self):
        return f'{self.title} ({self.manufacturer})'


class Brand(Assembler):
    extractor: type['VinDetails'] = None

    @property
    def brands(self) -> set['Brand']:
        return {self}


class UnsupportedBrand(Brand):
    """Unsupported brand."""


from ..common import constant_info
from ..dicts.bodies import BODY_MOTORCYCLE
from ._base import Detail, VinDetails


class DafraDetails(VinDetails):

    model = Detail(('vds', slice(0, 2)), {
        'CA': 'Speed 150',
        'CB': 'Kansas 150',
    })

    # for a while, it will be a motorcycle regardless of the code source
    body = Detail(None, constant_info(BODY_MOTORCYCLE))

    plant = Detail(('vis', 1), {
        'M': 'Manaus',
    })

    serial = Detail(('vis', slice(2, None)))


from ..dicts.bodies import *
from ._base import Detail, VinDetails


class RenaultDetails(VinDetails):
    """Renault VIN details extractor."""

    model = Detail(('vds', 1), {
        '0': 'Twingo',
        '1': 'R4',
        '2': 'R25',
        '3': 'R4',
        '4': ['R21', 'Express'],
        '5': ['Clio I', 'Laguna', 'R19', 'Safrane'],
        'A': ['Megane I', 'Master'],
        'B': 'Clio II',
        'C': 'Kangoo',
        'D': 'Master',
        'E': ['Espace III', 'Avantime'],
        'G': 'Laguna II',
        'H': 'Master Propulsion',
        'J': ['Vel Satis', 'New Trafic'],
        'K': 'Espace IV',
        'L': 'Trafic',
        'M': 'Megan II',
        'P': 'Modus',
        'S': ['Logan', 'Sandero', 'Duster', 'Dokker', 'Lodgy'],
        'Y': 'Koleos',
    })

    body = Detail(('vds', 0), {
        '2': BODY_SEDAN_2,
        '3': BODY_HATCH_3,
        '4': BODY_SEDAN_4,
        '5': BODY_HATCH_5,
        '6': BODY_SW_5,
        '7': BODY_CABRI_2,
        '8': BODY_COUPE_2,
        'A': BODY_SW_3,
        'B': BODY_HATCH_5,
        'C': BODY_HATCH_3,
        'D': BODY_COUPE_2,
        'E': BODY_CABRI_2,
        'F': BODY_VAN,
        'G': BODY_MINIVAN_3,
        'J': BODY_MINIVAN_5,
        'H': BODY_PICKUP_2,
        'K': BODY_SW_3,
        'L': BODY_SEDAN_4,
        'M': BODY_SEDAN_2,
        'N': BODY_MINIVAN_5,
        'S': BODY_SW_5,
        'U': BODY_PICKUP_2,
    })

    plant = Detail(('vds', 4), {
        'A': 'Portugal',
        'B': 'Batilly',
        'C': 'Creil',
        'D': 'Douai',
        'E': 'Spain',  # ? city
        'V': 'Spain',  # ? city
        'F': 'Flin',
        'G': 'Grand Coronne / Novo Mesto',
        'H': 'Haren',
        'J': 'Billancourt',
        'K': 'Dieppe',
        'N': 'Mexico',
        'P': 'Mexico',
        'R': 'Bursa / Moscow',
        'S': 'Sandouville',
        'T': 'Romorantin',
        'U': 'Maubeuge',
        'W': 'Valladolid',
        'X': 'Heuliez',
        'Z': 'USA',  # ? city
    })

    transmission = Detail(('vds', 5), {
        '1': 'Automatic, 3-Gears',
        '2': 'Automatic, 4-Gears',
        '4': 'Manual, 5-Gears',
        '5': 'Manual, 5-Gears',
        '8': 'Manual, 5-Gears, 4x4',
        'C': 'Manual, 5-Gears',
        'D': 'Manual, 5-Gears',
    })

    serial = Detail(('vis', slice(1, None)))



from typing import ClassVar

from .brands import Bajaj
from .brands import Dafra as DafraBrand
from .common import Assembler


class Dafra(Assembler):
    brands: ClassVar = {DafraBrand(), Bajaj(), 'BMW'}



import sys

import click

from vininfo import VERSION, Vin


@click.group()
@click.version_option(version=VERSION)
def entry_point():
    """vininfo command line utilities."""


@entry_point.command()
@click.argument('vin')
def show(vin):
    """Show information for VIN"""
    num = Vin(vin)

    click.secho('Basic:')

    def out(annotatable):
        for k, v in annotatable.annotate().items():
            click.secho(f'{k}: ', fg='green', nl=False)
            click.secho(v)

    out(num)

    details = num.details

    if details:
        click.secho('')
        click.secho('Details:')
        out(details)


@entry_point.command()
@click.argument('vin')
def check(vin):
    """Perform VIN checksum validation"""

    if Vin(vin).verify_checksum():
        click.secho('Checksum is valid', fg='green')
    else:
        click.secho('Checksum is not valid', fg='red', err=True)
        sys.exit(1)


def main():
    entry_point(obj={})


if __name__ == '__main__':
    main()



from ..common import candidate_by_year_model_mapping, constant_info
from ..dicts.bodies import BODY_MOTORCYCLE
from ._base import Detail, VinDetails


def get_wmi(details: VinDetails):
    # noinspection PyProtectedMember
    return details._vin.wmi


def get_years(details: VinDetails):
    # noinspection PyProtectedMember
    return details._vin.years


def get_model(details: VinDetails):
    """It looks like Bajaj Brazil is using the same VDS mapping that Bajaj India, or a pretty close one"""

    # Model name pattern with spaces
    bajaj_commons_mapping = {
        'A92': 'Dominar 160',
        'A36': 'Dominar 200',
        'A67': 'Dominar 400',
        'B65': 'Dominar 250',
    }

    bajaj_brazil_mapping_by_year_model = {
        '2026-': {
            'A92': 'Dominar NS160',
            'A36': 'Dominar NS200',
        },
        '-': {
            'C41': 'Pulsar N150'
        }
    }

    bajaj_brazil_mapping = bajaj_commons_mapping
    wmi = get_wmi(details)
    if wmi == '92T':
        bajaj_brazil_mapping = bajaj_commons_mapping.copy()
        updates = candidate_by_year_model_mapping(bajaj_brazil_mapping_by_year_model, get_years(details))
        if updates:
            bajaj_brazil_mapping.update(updates)

    candidates = {
        '92T': bajaj_brazil_mapping,
        '95V': {  # Bajaj assembled by Dafra
            '2A1': 'Dominar 160',
            '2B1': 'Dominar 200',
            '3B1': 'Dominar 400',
        },
        'MD2': {
            **bajaj_commons_mapping,
            # Bajaj India
            'B54': 'Pulsar N160',
            'B97': 'Pulsar N250',
        }
    }

    return candidates.get(wmi, {})


def get_plant(details):
    candidates = {
        'MD2': {  # Bajaj India
            'C': 'Chakan',
            'W': 'Waluj',  # Guessing
            'P': 'Pant Nagar'  # Guessing
        }
    }

    return candidates.get(get_wmi(details), {
        # WMI = 95V or 92T
        'M': 'Manaus'
    })


class BajajDetails(VinDetails):
    # for a while, it will be decoded by the first 3 characters of vds
    model = Detail(('vds', slice(0, 3)), get_model)

    # for a while, it will be a motorcycle regardless of the code source
    body = Detail(None, constant_info(BODY_MOTORCYCLE))

    plant = Detail(('vis', 1), get_plant)

    serial = Detail(('vis', slice(2, None)))



from .avtovaz import AvtoVazDetails
from .bajaj import BajajDetails
from .dafra import DafraDetails
from .nissan import NissanDetails
from .opel import OpelDetails
from .renault import RenaultDetails

__all__ = [
    'AvtoVazDetails', 'BajajDetails', 'DafraDetails', 'NissanDetails', 'OpelDetails', 'RenaultDetails'
]




import re
from datetime import datetime, timezone
from typing import TYPE_CHECKING, ClassVar

from .common import Annotatable, Assembler, Brand, UnsupportedBrand
from .dicts import COUNTRIES, REGIONS, WMI
from .exceptions import ValidationError

if TYPE_CHECKING:
    from .details._base import VinDetails


class Vin(Annotatable):
    """Offers basic VIN data extraction facilities."""

    annotate_titles: ClassVar = {
        'manufacturer': 'Manufacturer',
        'region': 'Region',
        'country': 'Country',
        'years': 'Years',
    }

    def __init__(self, num: str):
        self._brand = None
        self.num = self.validate(num)

        _details  = None
        for brand in self.assembler.brands:
            if isinstance(brand, str):
                brand = Brand(brand)
            details_extractor = brand.extractor

            if details_extractor:
                _details = details_extractor(self)
                if _details.model and _details.model.name:
                    self._brand = brand
                    break

        self.details: VinDetails = _details

    def __str__(self):
        return self.num

    @classmethod
    def validate(cls, num: str) -> str:
        """Performs basic VIN validation and sanation.

        :param num:

        """
        num = num.strip().upper()

        num_len = len(num)
        if num_len != 17:
            raise ValidationError(f'VIN number requires 17 chars ({num_len} given)')

        pattern = r"^[A-HJ-NPR-Z0-9]{17}$"
        if not re.match(pattern, num):
            raise ValidationError("VIN number must only contain alphanumeric symbols except 'I', 'O', and 'Q' ")

        return num

    def verify_checksum(self, *, check_year: bool = True) -> bool:
        """Performs checksum verification.

        .. warning:: Not every manufacturer uses VIN checksum rules.

        :param check_year: Whether to also check the model year.
            Note that not all manufacturer abey the rule. Default: True.

        """
        if check_year and self.vis[0] in {'U', 'Z', '0'}:
            return False

        trans = {
            'A': 1, 'B': 2, 'C': 3, 'D': 4, 'E': 5, 'F': 6, 'G': 7, 'H': 8,
            'J': 1, 'K': 2, 'L': 3, 'M': 4, 'N': 5,         'P': 7,         'R': 9,
                    'S': 2, 'T': 3, 'U': 4, 'V': 5, 'W': 6, 'X': 7, 'Y': 8, 'Z': 9,
        }
        weights = (8, 7, 6, 5, 4, 3, 2, 10, 0, 9, 8, 7, 6, 5, 4, 3, 2)

        checksum = 0

        for pos, char in enumerate(self.num):
            value = int(trans.get(char, char))
            checksum += (value * weights[pos])

        checksum = int(checksum) % 11

        check_digit = 'X' if checksum == 10 else checksum

        return f'{check_digit}' == self.vds[5]

    @property
    def wmi(self) -> str:
        """WMI (World Manufacturers Identification)"""
        return self.num[:3]

    @property
    def assembler(self) -> Assembler:
        """Assembler object."""

        wmi = self.wmi

        assembler = WMI.get(wmi)

        if not assembler:
            assembler = WMI.get(wmi[:2])

        if isinstance(assembler, str):
            assembler = Brand(assembler)

        if assembler is None:
            assembler = UnsupportedBrand()

        return assembler

    @property
    def brand(self) -> Brand:
        """Brand object."""
        brand = self._brand
        return UnsupportedBrand() if brand is None else brand

    @property
    def manufacturer(self) -> str:
        """Manufacturer title."""
        return self.assembler.manufacturer

    @property
    def manufacturer_is_small(self) -> bool:
        """A manufacturer who builds fewer than 1000 vehicles per year."""
        return f'{self.wmi[2]}' == '9'

    @property
    def vds(self) -> str:
        """VDS (Vehicle Descriptor Section)"""
        return self.num[3:9]

    @property
    def vis(self) -> str:
        """VIS (Vehicle Identifier Section)"""
        return self.num[9:17]

    @property
    def region_code(self) -> str:
        return self.wmi[0]

    @property
    def region(self) -> str | None:
        code = self.region_code

        title = None

        for chars, title_ in REGIONS.items():
            if code in chars:
                title = title_
                break

        return title

    @property
    def country_code(self) -> str:
        return self.wmi[0:2]

    @property
    def country(self) -> str | None:
        return COUNTRIES.get(self.country_code)

    @property
    def years_code(self) -> str:
        return self.vis[0]

    @property
    def years(self) -> list[int]:
        letters = 'ABCDEFGHJKLMNPRSTVWXY123456789'
        overflow_delta = len(letters)
        start_year_iso_table = 1980
        net_year = datetime.now(tz=timezone.utc).year + 1

        try:
            # E.g. .years_code == O (oh)
            delta = letters.index(self.years_code)
        except ValueError:
            return []

        year = delta + start_year_iso_table
        result = [year]
        while year + overflow_delta <= net_year:
            year += overflow_delta
            result.append(year)

        result.sort(reverse=True)

        return result

    @property
    def squish_vin(self) -> str:
        """Squish (or Pattern) VIN.

        The first 11 digits of the VIN minus the 9th digit (positions 1-8,
        positions 10 and 11).

        Squish VIN encodes vehicle information while preventing its precise
        identification. It can be useful for anonymization and privacy purposes.
        """
        return self.num[:8] + self.num[9:11]
